#!/usr/bin/env node
'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var fs = _interopDefault(require('fs'));
var path = _interopDefault(require('path'));
var os = _interopDefault(require('os'));

function __awaiter(thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

// Only Node.JS has a process variable that is of [[Class]] process
var detectNode = Object.prototype.toString.call(typeof process !== 'undefined' ? process : 0) === '[object process]';

const _GM = (typeof GM === 'object' ? GM : undefined);

const DISCORD_URL = 'https://discord.gg/gSsTUvJmD8';
const escapeFilename = (s) => {
    return s.replace(/[\s<>:{}"/\\|?*~.\0\cA-\cZ]+/g, '_');
};
const NODE_FETCH_HEADERS = {
    'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:81.0) Gecko/20100101 Firefox/81.0',
    'Accept-Language': 'en-US,en;q=0.8',
};
const getFetch = () => {
    if (!detectNode) {
        return fetch;
    }
    else {
        // eslint-disable-next-line @typescript-eslint/no-var-requires
        const nodeFetch = require('node-fetch');
        return (input, init) => {
            init = Object.assign({ headers: NODE_FETCH_HEADERS }, init);
            // eslint-disable-next-line @typescript-eslint/no-unsafe-return
            return nodeFetch(input, init);
        };
    }
};
const fetchData = (url, init) => __awaiter(void 0, void 0, void 0, function* () {
    const r = yield fetch(url, init);
    const data = yield r.arrayBuffer();
    return new Uint8Array(data);
});
const assertRes = (r) => {
    if (!r.ok)
        throw new Error(`${r.url} ${r.status} ${r.statusText}`);
};

const MSCZ_BUF_SYM = Symbol('msczBufferP');
const MSCZ_URL_SYM = Symbol('msczUrl');
const MAIN_CID_SYM = Symbol('mainCid');
const IPNS_KEY = 'QmSdXtvzC8v8iTTZuj5cVmiugnzbR1QATYRcGix4bBsioP';
const IPNS_RS_URL = `https://ipfs.io/api/v0/dag/resolve?arg=/ipns/${IPNS_KEY}`;
const getMainCid = (scoreinfo, _fetch = getFetch()) => __awaiter(void 0, void 0, void 0, function* () {
    // look for the persisted msczUrl inside scoreinfo
    let result = scoreinfo.store.get(MAIN_CID_SYM);
    if (result) {
        return result;
    }
    const r = yield _fetch(IPNS_RS_URL);
    assertRes(r);
    const json = yield r.json();
    result = json.Cid['/'];
    scoreinfo.store.set(MAIN_CID_SYM, result); // persist to scoreinfo
    return result;
});
const loadMsczUrl = (scoreinfo, _fetch = getFetch()) => __awaiter(void 0, void 0, void 0, function* () {
    // look for the persisted msczUrl inside scoreinfo
    let result = scoreinfo.store.get(MSCZ_URL_SYM);
    if (result) {
        return result;
    }
    const mainCid = yield getMainCid(scoreinfo, _fetch);
    const url = scoreinfo.getMsczCidUrl(mainCid);
    const r0 = yield _fetch(url);
    // ipfs-http-gateway specific error
    // may read further error msg as json
    if (r0.status !== 500) {
        assertRes(r0);
    }
    const cidRes = yield r0.json();
    const cid = cidRes.Key;
    if (!cid) {
        // read further error msg
        const err = cidRes.Message;
        if (err.includes('no link named')) { // file not found
            throw new Error('Score not in dataset');
        }
        else {
            throw new Error(err);
        }
    }
    result = `https://ipfs.infura.io/ipfs/${cid}`;
    scoreinfo.store.set(MSCZ_URL_SYM, result); // persist to scoreinfo
    return result;
});
const fetchMscz = (scoreinfo, _fetch = getFetch()) => __awaiter(void 0, void 0, void 0, function* () {
    let msczBufferP = scoreinfo.store.get(MSCZ_BUF_SYM);
    if (!msczBufferP) {
        msczBufferP = (() => __awaiter(void 0, void 0, void 0, function* () {
            const url = yield loadMsczUrl(scoreinfo, _fetch);
            const r = yield _fetch(url);
            assertRes(r);
            const data = yield r.arrayBuffer();
            return data;
        }))();
        scoreinfo.store.set(MSCZ_BUF_SYM, msczBufferP);
    }
    return msczBufferP;
});
// eslint-disable-next-line @typescript-eslint/require-await
const setMscz = (scoreinfo, buffer) => __awaiter(void 0, void 0, void 0, function* () {
    scoreinfo.store.set(MSCZ_BUF_SYM, Promise.resolve(buffer));
});

/**
 * type checking only so no missing keys
 */
function createLocale(obj) {
    return Object.freeze(obj);
}

var en = createLocale({
    'PROCESSING'() {
        return 'Processing…';
    },
    'BTN_ERROR'() {
        return '❌Download Failed!';
    },
    'DEPRECATION_NOTICE'(btnName) {
        return `DEPRECATED!\nUse \`${btnName}\` inside \`Individual Parts\` instead.\n(This may still work. Click \`OK\` to continue.)`;
    },
    'DOWNLOAD'(fileType) {
        return `Download ${fileType}`;
    },
    'DOWNLOAD_AUDIO'(fileType) {
        return `Download ${fileType} Audio`;
    },
    'IND_PARTS'() {
        return 'Individual Parts';
    },
    'IND_PARTS_TOOLTIP'() {
        return 'Download individual parts (BETA)';
    },
    'VIEW_IN_LIBRESCORE'() {
        return 'View in LibreScore';
    },
    'FULL_SCORE'() {
        return 'Full score';
    },
});

var es = createLocale({
    'PROCESSING'() {
        return 'Cargando…';
    },
    'BTN_ERROR'() {
        return '❌¡Descarga Fallida!';
    },
    'DEPRECATION_NOTICE'(btnName) {
        return `¡OBSOLETO!\nUtilizar \`${btnName}\` dentro de \`Partes Indivduales\` en su lugar.\n(Esto todavía puede funcionar. Pulsa \`Aceptar\` para continuar.)`;
    },
    'DOWNLOAD'(fileType) {
        return `Descargar ${fileType}`;
    },
    'DOWNLOAD_AUDIO'(fileType) {
        return `Descargar Audio ${fileType}`;
    },
    'IND_PARTS'() {
        return 'Partes individuales';
    },
    'IND_PARTS_TOOLTIP'() {
        return 'Descargar partes individuales (BETA)';
    },
    'VIEW_IN_LIBRESCORE'() {
        return 'Visualizar en LibreScore';
    },
    'FULL_SCORE'() {
        return 'Partitura Completa';
    },
});

var it = createLocale({
    'PROCESSING'() {
        return 'Caricamento…';
    },
    'BTN_ERROR'() {
        return '❌Download Fallito!';
    },
    'DEPRECATION_NOTICE'(btnName) {
        return `¡DEPRECATO!\nUtilizzare \`${btnName}\` all'interno di \`Parti Indivduali\`.\n(Qusto potrebbe funzionare. Cliccare \`Ok\` per continuare.)`;
    },
    'DOWNLOAD'(fileType) {
        return `Scaricare ${fileType}`;
    },
    'DOWNLOAD_AUDIO'(fileType) {
        return `Scaricare ${fileType} Audio`;
    },
    'IND_PARTS'() {
        return 'Parti Singole';
    },
    'IND_PARTS_TOOLTIP'() {
        return 'Scaricare Parti Singole (BETA)';
    },
    'VIEW_IN_LIBRESCORE'() {
        return 'Visualizzare in LibreScore';
    },
    'FULL_SCORE'() {
        return 'Spartito Completo';
    },
});

var zh = createLocale({
    'PROCESSING'() {
        return '处理中…';
    },
    'BTN_ERROR'() {
        return '❌下载失败!';
    },
    'DEPRECATION_NOTICE'(btnName) {
        return `不建议使用\n请使用 \`单独分谱\` 里的 \`${btnName}\` 按钮代替\n（这也许仍会起作用。单击\`确定\`以继续。）`;
    },
    'DOWNLOAD'(fileType) {
        return `下载 ${fileType}`;
    },
    'DOWNLOAD_AUDIO'(fileType) {
        return `下载 ${fileType} 音频`;
    },
    'IND_PARTS'() {
        return '单独分谱';
    },
    'IND_PARTS_TOOLTIP'() {
        return '下载单独分谱 (BETA)';
    },
    'VIEW_IN_LIBRESCORE'() {
        return '在 LibreScore 中查看';
    },
    'FULL_SCORE'() {
        return '完整乐谱';
    },
});

const locales = ((l) => Object.freeze(l))({
    en,
    es,
    it,
    zh,
});
// detect browser language
const lang = (() => {
    let userLangs;
    if (!detectNode) {
        userLangs = navigator.languages;
    }
    else {
        const env = process.env;
        const l = env.LC_ALL || env.LC_MESSAGES || env.LANG || env.LANGUAGE || '';
        userLangs = [l.slice(0, 2)];
    }
    const names = Object.keys(locales);
    const _lang = userLangs.find(l => {
        // find the first occurrence of valid languages
        return names.includes(l);
    });
    return _lang || 'en';
})();
// eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
function i18n(key) {
    const locale = locales[lang];
    return locale[key];
}

var name = "musescore-downloader";
var version = "0.23.15";
var dependencies = {
	"@librescore/fonts": "^0.4.0",
	"@librescore/sf3": "^0.3.0",
	"detect-node": "^2.0.4",
	inquirer: "^7.3.3",
	"node-fetch": "^2.6.1",
	ora: "^5.1.0",
	webmscore: "^0.18.0"
};

/* eslint-disable @typescript-eslint/no-var-requires */
const WEBMSCORE_URL = `https://cdn.jsdelivr.net/npm/webmscore@${dependencies.webmscore}/webmscore.js`;
// fonts for Chinese characters (CN) and Korean hangul (KR)
// JP characters are included in the CN font
const FONT_URLS = ['CN', 'KR'].map(l => `https://cdn.jsdelivr.net/npm/@librescore/fonts@${dependencies['@librescore/fonts']}/SourceHanSans${l}.min.woff2`);
const SF3_URL = `https://cdn.jsdelivr.net/npm/@librescore/sf3@${dependencies['@librescore/sf3']}/FluidR3Mono_GM.sf3`;
const SOUND_FONT_LOADED = Symbol('SoundFont loaded');
const initMscore = (w) => __awaiter(void 0, void 0, void 0, function* () {
    if (!detectNode) { // attached to a page
        if (!w['WebMscore']) {
            // init webmscore (https://github.com/LibreScore/webmscore)
            const script = w.document.createElement('script');
            script.src = WEBMSCORE_URL;
            w.document.body.append(script);
            yield new Promise(resolve => { script.onload = resolve; });
        }
        return w['WebMscore'];
    }
    else { // nodejs
        return require('webmscore').default;
    }
});
let fonts;
const initFonts = () => {
    // load CJK fonts
    // CJK (East Asian) characters will be rendered as "tofu" if there is no font
    if (!fonts) {
        if (detectNode) {
            // module.exports.CN = ..., module.exports.KR = ...
            const FONTS = Object.values(require('@librescore/fonts'));
            const fs = require('fs');
            fonts = Promise.all(FONTS.map((path) => fs.promises.readFile(path)));
        }
        else {
            fonts = Promise.all(FONT_URLS.map(url => fetchData(url)));
        }
    }
};
const loadSoundFont = (score) => {
    if (!score[SOUND_FONT_LOADED]) {
        const loadPromise = (() => __awaiter(void 0, void 0, void 0, function* () {
            let data;
            if (detectNode) {
                // module.exports.FluidR3Mono = ...
                const SF3 = Object.values(require('@librescore/sf3'))[0];
                const fs = require('fs');
                data = yield fs.promises.readFile(SF3);
            }
            else {
                data = yield fetchData(SF3_URL);
            }
            yield score.setSoundFont(data);
        }))();
        score[SOUND_FONT_LOADED] = loadPromise;
    }
    return score[SOUND_FONT_LOADED];
};
const loadMscore = (scoreinfo, w) => __awaiter(void 0, void 0, void 0, function* () {
    initFonts();
    const WebMscore = yield initMscore(w);
    // parse mscz data
    const data = new Uint8Array(new Uint8Array(yield fetchMscz(scoreinfo)));
    const score = yield WebMscore.load('mscz', data, yield fonts);
    yield score.generateExcerpts();
    return score;
});
const INDV_DOWNLOADS = [
    {
        name: i18n('DOWNLOAD')('PDF'),
        fileExt: 'pdf',
        action: (score) => score.savePdf(),
    },
    {
        name: i18n('DOWNLOAD')('MSCZ'),
        fileExt: 'mscz',
        action: (score) => score.saveMsc('mscz'),
    },
    {
        name: i18n('DOWNLOAD')('MusicXML'),
        fileExt: 'mxl',
        action: (score) => score.saveMxl(),
    },
    {
        name: i18n('DOWNLOAD')('MIDI'),
        fileExt: 'mid',
        action: (score) => score.saveMidi(true, true),
    },
    {
        name: i18n('DOWNLOAD_AUDIO')('MP3'),
        fileExt: 'mp3',
        action: (score) => loadSoundFont(score).then(() => score.saveAudio('mp3')),
    },
    {
        name: i18n('DOWNLOAD_AUDIO')('FLAC'),
        fileExt: 'flac',
        action: (score) => loadSoundFont(score).then(() => score.saveAudio('flac')),
    },
    {
        name: i18n('DOWNLOAD_AUDIO')('OGG'),
        fileExt: 'ogg',
        action: (score) => loadSoundFont(score).then(() => score.saveAudio('ogg')),
    },
];

class ScoreInfo {
    constructor() {
        this.RADIX = 20;
        this.INDEX_RADIX = 32;
        this.store = new Map();
    }
    get idLastDigit() {
        return (+this.id) % this.RADIX;
    }
    get fileName() {
        return escapeFilename(this.title);
    }
    getMsczIpfsRef(mainCid) {
        return `/ipfs/${mainCid}/${this.idLastDigit}/${this.id}.mscz`;
    }
    getMsczCidUrl(mainCid) {
        return `https://ipfs.infura.io:5001/api/v0/block/stat?arg=${this.getMsczIpfsRef(mainCid)}`;
    }
    getScorepackRef(mainCid) {
        return `/ipfs/${mainCid}/index/${(+this.id) % this.INDEX_RADIX}/${this.id}`;
    }
}
class ScoreInfoObj extends ScoreInfo {
    constructor(id = 0, title = '') {
        super();
        this.id = id;
        this.title = title;
    }
}
class ScoreInfoHtml extends ScoreInfo {
    constructor(html) {
        super();
        this.html = html;
        this.ID_REG = /<meta property="al:ios:url" content="musescore:\/\/score\/(\d+)">/;
        this.TITLE_REG = /<meta property="og:title" content="(.*)">/;
        this.BASEURL_REG = /<meta property="og:image" content="(.+\/)score_.*">/;
    }
    get id() {
        const m = this.html.match(this.ID_REG);
        if (!m)
            return 0;
        return +m[1];
    }
    get title() {
        const m = this.html.match(this.TITLE_REG);
        if (!m)
            return '';
        return m[1];
    }
    get baseUrl() {
        const m = this.html.match(this.BASEURL_REG);
        if (!m)
            return '';
        return m[1];
    }
    static request(url, _fetch = getFetch()) {
        return __awaiter(this, void 0, void 0, function* () {
            const r = yield _fetch(url);
            if (!r.ok)
                return new ScoreInfoHtml('');
            const html = yield r.text();
            return new ScoreInfoHtml(html);
        });
    }
}
const getActualId = (scoreinfo, _fetch = getFetch()) => __awaiter(void 0, void 0, void 0, function* () {
    if (scoreinfo.id <= 1000000000000) {
        // actual id already
        return scoreinfo.id;
    }
    const mainCid = yield getMainCid(scoreinfo, _fetch);
    const ref = `${mainCid}/sid2id/${scoreinfo.id}`;
    const url = `https://ipfs.infura.io:5001/api/v0/dag/get?arg=${ref}`;
    const r0 = yield _fetch(url);
    if (r0.status !== 500) {
        assertRes(r0);
    }
    const res = yield r0.json();
    if (typeof res !== 'number') {
        // read further error msg
        throw new Error(res.Message);
    }
    // assign the actual id back to scoreinfo
    Object.defineProperty(scoreinfo, 'id', {
        get() { return res; },
    });
    return res;
});

const _getLink = (indexingInfo) => {
    const { scorepack } = JSON.parse(indexingInfo);
    return `https://librescore.org/score/${scorepack}`;
};
const getLibreScoreLink = (scoreinfo, _fetch = getFetch()) => __awaiter(void 0, void 0, void 0, function* () {
    const mainCid = yield getMainCid(scoreinfo, _fetch);
    const ref = scoreinfo.getScorepackRef(mainCid);
    const url = `https://ipfs.infura.io:5001/api/v0/dag/get?arg=${ref}`;
    const r0 = yield _fetch(url);
    if (r0.status !== 500) {
        assertRes(r0);
    }
    const res = yield r0.json();
    if (typeof res !== 'string') {
        // read further error msg
        throw new Error(res.Message);
    }
    return _getLink(res);
});

/* eslint-disable @typescript-eslint/explicit-module-boundary-types */
const IS_NPX_REG = /_npx(\/|\\)\d+\1/;
const NPM_REGISTRY = 'https://registry.npmjs.org';
function isNpx() {
    // file is in a npx cache dir
    // TODO: installed locally?
    return __dirname.match(IS_NPX_REG) !== null;
}
function getSelfVer() {
    return version;
}
function getLatestVer(_fetch = getFetch()) {
    return __awaiter(this, void 0, void 0, function* () {
        // fetch pkg info from the npm registry
        const r = yield _fetch(`${NPM_REGISTRY}/${name}`);
        const json = yield r.json();
        return json['dist-tags'].latest;
    });
}
function getVerInfo() {
    return __awaiter(this, void 0, void 0, function* () {
        const installed = getSelfVer();
        const latest = yield getLatestVer();
        return {
            installed,
            latest,
            isLatest: installed === latest,
        };
    });
}

/* eslint-disable @typescript-eslint/no-var-requires */
const inquirer = require('inquirer');
const ora = require('ora');
const chalk = require('chalk');
const SCORE_URL_PREFIX = 'https://(s.)musescore.com/';
const SCORE_URL_REG = /https:\/\/(s\.)?musescore\.com\//;
const EXT = '.mscz';
void (() => __awaiter(void 0, void 0, void 0, function* () {
    const arg = process.argv[2];
    if (['-v', '--version'].includes(arg)) { // ran with flag -v or --version, `msdl -v`
        console.log(getSelfVer()); // print musescore-downloader version
        return; // exit process
    }
    // Determine platform and paste message
    const platform = os.platform();
    let pasteMessage = '';
    if (platform === 'win32') {
        pasteMessage = 'right-click to paste';
    }
    else if (platform === 'linux') {
        pasteMessage = 'usually Ctrl+Shift+V to paste';
    } // For MacOS, no hint is needed because the paste shortcut is universal.
    let scoreinfo;
    let librescoreLink;
    // ask for the page url or path to local file
    const { fileInit } = yield inquirer.prompt({
        type: 'input',
        name: 'fileInit',
        message: 'Score URL or path to local MSCZ file:',
        suffix: '\n  ' +
            `(starts with "${SCORE_URL_PREFIX}" or local filepath ends with "${EXT}") ` +
            `${chalk.bgGray(pasteMessage)}\n `,
        validate(input) {
            return input &&
                (!!input.match(SCORE_URL_REG) ||
                    (input.endsWith(EXT) && fs.statSync(input).isFile()));
        },
        default: arg,
    });
    const isLocalFile = fileInit.endsWith(EXT);
    if (!isLocalFile) {
        // request scoreinfo
        scoreinfo = yield ScoreInfoHtml.request(fileInit);
        try {
            yield getActualId(scoreinfo);
        }
        catch (err) {
            console.error(err);
        }
        // confirmation
        const { confirmed } = yield inquirer.prompt({
            type: 'confirm',
            name: 'confirmed',
            message: 'Continue?',
            prefix: `${chalk.yellow('!')} ` +
                `ID: ${scoreinfo.id}\n  ` +
                `Title: ${scoreinfo.title}\n `,
            default: true,
        });
        if (!confirmed)
            return;
        // initiate LibreScore link request
        librescoreLink = getLibreScoreLink(scoreinfo);
        librescoreLink.catch(() => ''); // silence this unhandled Promise rejection
        // print a blank line to the terminal
        console.log();
    }
    else {
        scoreinfo = new ScoreInfoObj(0, path.basename(fileInit, EXT));
    }
    const spinner = ora({
        text: i18n('PROCESSING')(),
        color: 'blue',
        spinner: 'bounce',
        indent: 0,
    }).start();
    let score;
    let metadata;
    try {
        if (!isLocalFile) {
            // fetch mscz file from the dataset, and cache it for side effect
            yield fetchMscz(scoreinfo);
        }
        else {
            // load local file
            const data = yield fs.promises.readFile(fileInit);
            yield setMscz(scoreinfo, data.buffer);
        }
        spinner.info('MSCZ file loaded');
        if (!isLocalFile) {
            spinner.info(`File URL: ${scoreinfo.store.get(MSCZ_URL_SYM)}`);
        }
        if (librescoreLink) {
            try {
                spinner.info(`${i18n('VIEW_IN_LIBRESCORE')()}: ${yield librescoreLink}`);
            }
            catch (_a) { } // it doesn't affect the main feature
        }
        spinner.start();
        // load score using webmscore
        score = yield loadMscore(scoreinfo);
        metadata = yield score.metadata();
        spinner.info('Score loaded by webmscore');
    }
    catch (err) {
        spinner.fail(err.message);
        spinner.info('Send your URL to the #dataset-bugs channel in the LibreScore Community Discord server:\n  ' +
            DISCORD_URL);
        return;
    }
    spinner.succeed('OK\n');
    // build part choices
    const partChoices = metadata.excerpts.map(p => ({ name: p.title, value: p.id }));
    // add the "full score" option as a "part" 
    partChoices.unshift({ value: -1, name: i18n('FULL_SCORE')() });
    // build filetype choices
    const typeChoices = INDV_DOWNLOADS.map((d, i) => ({ name: d.name, value: i }));
    // part selection
    const { part } = yield inquirer.prompt({
        type: 'list',
        name: 'part',
        message: 'Part Selection',
        choices: partChoices,
    });
    const partName = partChoices[part + 1].name;
    yield score.setExcerptId(part);
    // filetype selection
    const { types } = yield inquirer.prompt({
        type: 'checkbox',
        name: 'types',
        message: 'Filetype Selection',
        choices: typeChoices,
        validate(input) {
            return input.length >= 1;
        },
    });
    const filetypes = types.map(i => INDV_DOWNLOADS[i]);
    // destination directory
    const { dest } = yield inquirer.prompt({
        type: 'input',
        name: 'dest',
        message: 'Destination Directory:',
        validate(input) {
            return input && fs.statSync(input).isDirectory();
        },
        default: process.cwd(),
    });
    // export files
    const fileName = scoreinfo.fileName || (yield score.titleFilenameSafe());
    spinner.start();
    yield Promise.all(filetypes.map((d) => __awaiter(void 0, void 0, void 0, function* () {
        const data = yield d.action(score);
        const n = `${fileName} - ${escapeFilename(partName)}.${d.fileExt}`;
        const f = path.join(dest, n);
        yield fs.promises.writeFile(f, data);
        spinner.info(`Saved ${chalk.underline(f)}`);
        spinner.start();
    })));
    spinner.succeed('OK');
    if (!isNpx()) {
        const { installed, latest, isLatest } = yield getVerInfo();
        if (!isLatest) {
            console.log(chalk.yellowBright(`\nYour installed version (${installed}) of the musescore-downloader CLI is not the latest one (${latest})!\nRun npm i -g musescore-downloader@${latest} to update.`));
        }
    }
}))();
